open B8_dump
open B8_prop
open B8_types
open B8_utils

let debug_merge = false
let debug_flat = false
  
(********************************************************************

SIMPLIFY:

- Remove Tlist with one sub-doc
  
  ********************************************************************)
  
let not_found = Not_found

  
let only_spaces s =
  try
    for i = 0 to String.length s - 1 do
      if not (is_space s.[i]) then raise not_found
    done;
    true
  with _  -> false

    (* spc = 2: simple space, spc = 1: underline space *)

let rec simplify tdoc =
  match tdoc.expr with
    Ttag_B tdoc2 -> 
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc > 0 then recv else 0, { tdoc with expr = Ttag_B tdoc2 }
        
  | Ttag_EM tdoc2 ->
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc > 0 then recv else 0, { tdoc with expr = Ttag_EM tdoc2 }
  
  | Ttag_I tdoc2 -> 
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc > 0 then recv else 0, { tdoc with expr = Ttag_I tdoc2 }
  
  | Ttag_PL tdoc2 -> 
      let (spc, tdoc2) as recv = simplify tdoc2 in
      spc, { tdoc with expr = Ttag_PL tdoc2 }

  | Ttag_S tdoc2 ->
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc > 0 then recv else 0, { tdoc with expr = Ttag_S tdoc2 }
          
  | Ttag_TT tdoc2 -> 
      let (spc, tdoc2) as recv = simplify tdoc2 in
      0, { tdoc with expr = Ttag_TT tdoc2 }
        
  | Ttag_U tdoc2 -> 
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc < 2 then spc, { tdoc with expr = Ttag_U tdoc2 } else recv

  | Ttag_color (tdoc2, color)  ->  
      let (spc, tdoc2) as recv = simplify tdoc2 in
      if spc < 2 then spc, { tdoc with expr = Ttag_color (tdoc2,color) } 
        else recv

  | Ttag_size (tdoc2, size) ->  
      let (spc, tdoc2) as recv = simplify tdoc2 in
      spc, { tdoc with expr = Ttag_size (tdoc2, size) }
      
  | Tlist [tdoc2] -> simplify tdoc2
  | Tlist list -> 
      let spc, list = List.fold_left (fun (spc, list) tdoc2 -> 
            let spc2, tdoc2 = simplify tdoc2 in
            (min spc spc2), tdoc2 :: list) (2,[]) list in
      spc, { tdoc with expr = Tlist (List.rev list) }

  | Tstring s -> 
      (if only_spaces s then
        if tdoc.prop.u > 0 then 1 else 2 else 0), tdoc

let simplify tdoc =
  let _, tdoc = simplify tdoc in
  tdoc
  
  (********************************************************************

MERGE:

- Merge adjacent blocks if same attributes.
- Whitespace elimination.
  
  ********************************************************************)

let rec remove_spc s1 s2 pos1 pos2 len =
  if debug_merge then begin
  Printf.printf "[%s][%s] %d %d %d" (String.escaped s1)
  (String.escaped s2) pos1 pos2 len;
print_newline ();
  end;
  let p2 = pos2+1 in
  if pos1 >= len then 
    if p2 = len then s2
    else begin
        let s = String.sub s2 0 p2 in
        if debug_merge then begin
Printf.printf "SUB[%d]" p2; print_newline (); 
end;
        s
      end
  else
  let c2 = s2.[pos2] in
  let c1 = s1.[pos1] in
  if is_space c2 && is_space c1 then 
    remove_spc s1 s2 (pos1+1) pos2 len
  else
    begin
      if debug_merge then begin
Printf.printf "COPY %d" p2; print_newline (); 
end;
      s2.[p2] <- c1;
      remove_spc s1 s2 (pos1+1) p2 len
    end

let remove_spaces s pos =
  if debug_merge then begin
Printf.printf "[%s] >>>>" (String.escaped s); print_newline (); 
  end;
  let len = String.length s in
  let s = if len > 1 then
      let s2 = String.create len in
      String.blit s 0 s2 0 pos;
      s2.[pos] <- s.[pos];
      let s = remove_spc s s2 pos (pos-1) len in
      s
    else s in
  if debug_merge then begin
      Printf.printf ">>>>> [%s]" (String.escaped s); print_newline ();
    end;
  s
  
let rec iter_merge tdoc last =
  match tdoc.expr with
    Ttag_B tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_B tdoc2 } in
      last, tdoc
  | Ttag_EM tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_EM tdoc2 } in
      last, tdoc
  | Ttag_I tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_I tdoc2 } in
      last, tdoc
  | Ttag_PL tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_PL tdoc2 } in
      last, tdoc
  | Ttag_S tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_S tdoc2 } in
      last, tdoc
  | Ttag_TT tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_TT tdoc2 } in
      last, tdoc
  | Ttag_U tdoc2 -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_U tdoc2 } in
      last, tdoc
  | Ttag_color (tdoc2, color)  -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_color (tdoc2, color) } in
      last, tdoc
  | Ttag_size (tdoc2, size) -> let (last, tdoc2) = iter_merge tdoc2 last in
      let tdoc = { tdoc with expr = Ttag_size (tdoc2, size) } in
      last, tdoc
  | Tlist list -> 
      let last, list = List.fold_left (fun (last, list) tdoc ->
            try
              let last, tdoc = iter_merge tdoc last in
              (last, tdoc :: list)
            with _ ->
                last, list
        ) (last, []) list in
      if list = [] then raise not_found else
        last, { tdoc with expr = Tlist (List.rev list) }
  | Tstring "" -> raise not_found
  | Tstring s -> 
      if debug_merge then begin
          Printf.printf "STRING {%s}" (String.escaped s); print_newline (); 
        end;
      let prop = tdoc.prop in
      match last with
        None -> let tdoc = { tdoc with expr = Tstring 
                (if prop.tt then s else remove_spaces s 1) } in
          (Some tdoc), tdoc
      | Some ({ expr = Tstring s2 } as tdoc2) ->
          let prop2 = tdoc2.prop in
          if prop = prop2 ||
            ((only_spaces s (*|| only_spaces s2*)) && 
              space_prop prop = space_prop prop2)
          then
            ( 
              if debug_merge then begin
                  Printf.printf "Same prop"; print_newline ();
                end;
              tdoc2.expr <- Tstring (
                if prop.tt then s2 ^ s else
                  remove_spaces (s2 ^ s) (String.length s2)); 
              if only_spaces s2 then tdoc2.prop <- prop;
              raise not_found)
          else
          if only_spaces s2 && space_prop prop = space_prop prop2 then
            begin
              tdoc2.expr <- Tstring "";
              let tdoc = { tdoc with expr = Tstring 
                    (if prop.tt then 
                      s2 ^ s else remove_spaces (s2 ^s) (String.length s2)) } in
              (Some tdoc), tdoc
            end
          else
          let len2 = String.length s2 in
          let s = if prop.tt then s else 
            let s = remove_spaces s 1 in
            if is_space s.[0] &&
              space_prop prop = space_prop prop2 &&
              is_space s2.[len2-1] then begin
                if debug_merge then begin
                    Printf.printf "Remove leading space"; print_newline ();
                  end;
                let len = String.length s in
                if len = 1 then raise not_found else
                  String.sub s 1 (len-1)
              end else s
          in
          let tdoc = { tdoc with expr = Tstring s } in
          (Some tdoc), tdoc
      | _ -> assert false
          
let merge tdoc =
  try
    let _, tdoc = iter_merge tdoc None in
    simplify tdoc
  with _ -> { expr = Tlist [] ; prop = prop_root }

        
(********************************************************************

FLATTEN:

- Build a document from a flattened version 
- No Whitespace elimination (must be done before).
  
  ********************************************************************)

      (* bit correspondance *)
let b_bit = 0
let em_bit = 1
let i_bit = 2
let s_bit = 3
let tt_bit = 4

let u_bit = 5
let uu_bit = 6
let uuu_bit = 7

let s0_bit = 8
let s1_bit = 9
let s2_bit = 10
let s3_bit = 11
let s4_bit = 12
let s5_bit = 13
let s6_bit = 14
let s7_bit = 15
let s8_bit = 16
let s9_bit = 17

let cr_bit = 18
let cg_bit = 19
let cb_bit = 20
let cc_bit = 21
let cm_bit = 22
let cy_bit = 23
let ck_bit = 24
let cw_bit = 25
  
let pl_bit = 30   (* in a <PL>...</PL> range *)
let spc_bit = 29  (* a string containing only spaces (TT or not) *)
let uspc_bit = 28 (* a string containing only underline spaces *)

let ok_bit = 27
     (* int correspondance *)  
let b_int = 1 lsl 0
let em_int = 1 lsl 1
let i_int = 1 lsl 2
let s_int = 1 lsl 3
let tt_int = 1 lsl 4

let u_int = 1 lsl 5
let uu_int = 1 lsl 6
let uuu_int = 1 lsl 7

let s0_int = 1 lsl 8
let s1_int = 1 lsl 9
let s2_int = 1 lsl 10
let s3_int = 1 lsl 11
let s4_int = 1 lsl 12
let s5_int = 1 lsl 13
let s6_int = 1 lsl 14
let s7_int = 1 lsl 15
let s8_int = 1 lsl 16
let s9_int = 1 lsl 17

let cr_int = 1 lsl 18
let cg_int = 1 lsl 19
let cb_int = 1 lsl 20
let cc_int = 1 lsl 21
let cm_int = 1 lsl 22
let cy_int = 1 lsl 23
let ck_int = 1 lsl 24
let cw_int = 1 lsl 25
  
let pl_int = 1 lsl 30
let spc_int = 1 lsl 29
let uspc_int = 1 lsl 28
let ok_int = 1 lsl 27

  (*
let _ =
  Printf.printf "spc: %d uspc_int: %d ok_int: %d" 
    spc_int uspc_int ok_int; print_newline ()
  *)

let int_of_color color =
  match color with
    Red -> cr_int
  | Green -> cg_int
  | Blue -> cb_int
  | Cyan -> cc_int
  | Magenta -> cm_int
  | Yellow -> cy_int
  | Black -> ck_int
  | White -> cw_int
  | Unspecified -> 0

let int_of_size size =
  match size with
    T0 -> s0_int
  | T1 -> s1_int
  | T2 -> s2_int
  | T3 -> s3_int
  | T4 -> s4_int
  | T5 -> s5_int
  | T6 -> s6_int
  | T7 -> s7_int
  | T8 -> s8_int
  | T9 -> s9_int
  | Tnone -> 0

let color_of_int int =
  match int with
    1 -> Red
  | 2 -> Green
  | 4 -> Blue
  | 8 -> Cyan
  | 16 -> Magenta
  | 32 -> Yellow
  | 64 -> Black
  | 128 -> White
  | _ -> Unspecified

let size_of_int size =
  match size with
    1 -> T0
  | 2 -> T1
  | 4 -> T2
  | 8 -> T3
  | 16 -> T4
  | 32 -> T5
  | 64 -> T6
  | 128 -> T7
  | 256 -> T8
  | 512 -> T9
  | _ -> Tnone
  
let int_of_prop prop =
  (if prop.b then b_int else 0) +
    (if prop.em then em_int else 0) +
    (if prop.i then i_int else 0) +
    (if prop.s then s_int else 0) +
    (if prop.tt then tt_int else 0) +
    (if prop.u = 0 then 0 else
    if prop.u = 1 then u_int else
    if prop.u = 2 then u_int + uu_int else 
      u_int + uu_int + uuu_int) +
    (int_of_size prop.size) +
    (int_of_color prop.color) +
  ok_int

let sizes = s0_int + s1_int + s2_int + s3_int + s4_int + s5_int + s6_int +
    s7_int + s8_int + s9_int
let colors = cr_int + cg_int + cb_int + cc_int + cm_int + cy_int + ck_int +
    cw_int
  
let prop_of_int int =
  let uuu = (int lsr u_bit) land 7 in 
  {
    b = int land b_int <> 0;
    em = int land em_int <> 0;
    i = int land i_int <> 0;
    s = int land s_int <> 0;
    tt = int land tt_int <> 0;
    u = (match uuu with 1 -> 1| 3 -> 2 | 7 -> 3 | _ -> 0);
    size = size_of_int ((int land sizes) lsr s0_bit);
    color = color_of_int ((int land colors) lsr cr_bit);
  }

let is_spc p =
  p land (spc_int + uspc_int) <> 0
let is_u p =
  p land u_int <> 0
let space_int p =
  if is_u p then
    p land (lnot (s_int + em_int + i_int + b_int + spc_int + uspc_int))
  else
    p land (lnot (s_int + em_int + i_int + b_int + spc_int + uspc_int + colors))

type flat = {
    strs : string array;
    ints : int array;
    len : int;
    pos : int;
  }
    
  
let flatten_tdoc tdoc =
  let str_tab = ref [||] in
  let int_tab = ref [||] in
  let tab_len = ref 0 in
  let tab_pos = ref 0 in
  
  let rec iter tdoc =
    match tdoc.expr with
      Ttag_B tdoc
    | Ttag_EM tdoc
    | Ttag_I tdoc
    | Ttag_PL tdoc
    | Ttag_S tdoc
    | Ttag_TT tdoc
    | Ttag_U tdoc
    | Ttag_color (tdoc, _)
    | Ttag_size (tdoc, _) -> iter tdoc
    | Tlist list -> List.iter iter list
    | Tstring "" -> ()
    | Tstring string -> 
        let prop = int_of_prop tdoc.prop in
        let prop = if only_spaces string then begin
              if tdoc.prop.u = 0 then
                (space_int prop) + spc_int
              else
                (space_int prop) + uspc_int
            end
          else
            prop
        in
(*        Printf.printf "pos: %d len:%d" !tab_pos !tab_len; print_newline ();*)
        if !tab_pos = !tab_len then
          let new_len = !tab_len * 2 + 50 in
          let new_str_tab = Parray.create new_len string in
          let new_int_tab = Parray.create new_len prop in
          Parray.blit !str_tab 0 new_str_tab 0 !tab_pos;
          Parray.blit !int_tab 0 new_int_tab 0 !tab_pos;
          int_tab := new_int_tab;
          str_tab := new_str_tab;
          tab_len := new_len;
          incr tab_pos
        else
        let pos = !tab_pos in
        (!str_tab).(pos) <- string;
        (!int_tab).(pos) <- prop;
        tab_pos := pos + 1
  in
  iter tdoc;
  { strs = !str_tab;
    ints = !int_tab;
    len = !tab_pos;
    pos = 0;}
  
let list_tdoc list p = 
  {
    expr = Tlist list;
    prop = prop_of_int p;
  }

let string_tdoc s p =
  { expr = Tstring s;
    prop = prop_of_int p;
  }
  
type tag = {
    constr : int -> tdoc -> tdoc;
    up : bool; (* up: 0 in base, 1 in prop *)
    int : int; (* 1 lsl bit *)
    at_least : int; (* 1 bits to continue *)
    on_force_stop : int; (* 1 bits must stop *)
    base_shared : int;
    change_base : int -> int;
    name : string;
  }

let rec iter_nbits v n =
  if v = 0 then n else
  iter_nbits (v lsr 1) (n+(v land 1))
  
let nbits v =
  if v = 0 then 0 else
    iter_nbits v 0

let pl_ints =
  u_int lor uu_int lor uuu_int lor em_int lor s_int lor i_int lor b_int lor tt_int
  
(* detect PL optimization *)
let rec find_pl_place t pos =
  if pos < t.len then
    let p1 = t.ints.(pos-1) in
    let p2 = t.ints.(pos) in
    let diff = p1 lxor p2 in
    let diff_up = diff land p2 in
    let diff_down = diff land p1 in
    let p1_pl = p1 land (lnot pl_ints) in
    let diff_pl = p1_pl lxor p2 in
    let diff_up_pl = diff_pl land p2 in
    let win = nbits (diff_down land pl_ints) in
    let lose = nbits (diff_up_pl land pl_ints) in
    if debug_flat then begin
        Printf.printf "%d: win %d lose %d" pos win lose; print_newline ();
      end;
    if diff_up = 0 && win - lose > 2 then begin
        if debug_flat then begin
            Printf.printf "PL diff %d" (nbits (diff_down land pl_ints)); 
            print_newline ();
          end;
        find_end_pl t pos (pos+1) p1 p2
      end else
      find_pl_place t (pos+1)

and find_end_pl t pos pos2 p1 p2 =
  if pos2 = t.len then
    find_pl_place t (pos+1)
  else
  if t.ints.(pos2-1) = p2 &&
    t.ints.(pos2) = p1 then begin
      if debug_flat then begin
          Printf.printf "PL from %d to %d" pos pos2; print_newline ();
        end;
      for i = pos to pos2 - 1 do
        if debug_flat then begin
            Printf.printf "Marking %d PL" i; print_newline ();
          end;
        t.ints.(i) <- t.ints.(i) lor pl_int
      done;
      find_pl_place t (pos2+1)
    end
  else
    find_end_pl t pos (pos2+1) p1 p2

let mark_pl t pos = 
  if t.len - pos > 3 then
    find_pl_place t (pos+1) 

let b_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_B tdoc; prop = prop_of_int base });
    up = true;
    int = b_int;
    at_least = b_int lor spc_int lor uspc_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor b_int);
    name = "<B>";
  }

let i_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_I tdoc; prop = prop_of_int base });
    up = true;
    int = i_int;
    at_least = i_int lor spc_int lor uspc_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor i_int);
    name = "<I>";
  }

let em_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_EM tdoc; prop = prop_of_int base });
    up = true;
    int = em_int;
    at_least = em_int lor s_int lor spc_int lor uspc_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor em_int);
    name = "<EM>";
  }

let s_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_S tdoc; prop = prop_of_int base });
    up = true;
    int = s_int;
    at_least = s_int lor spc_int lor uspc_int lor pl_int;
    on_force_stop = em_int;
    base_shared = 0;
    change_base = (fun base -> (base lor s_int) land (lnot em_int));
    name = "<S>";
  }

let tt_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_TT tdoc; prop = prop_of_int base });
    up = true;
    int = tt_int;
    at_least = tt_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor tt_int);
    name = "<TT>";
  }

let u_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_U tdoc; prop = prop_of_int base });
    up = true;
    int = u_int;
    at_least = u_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor u_int);
    name = "<U>";
  }

let uu_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_U tdoc; prop = prop_of_int base });
    up = true;
    int = uu_int;
    at_least = uu_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor uu_int);
    name = "<UU>";
  }

let uuu_tag = {
    constr = (fun base tdoc -> 
        { expr = Ttag_U tdoc; prop = prop_of_int base });
    up = true;
    int = uuu_int;
    at_least = uuu_int lor pl_int;
    on_force_stop = 0;
    base_shared = 0;
    change_base = (fun base -> base lor uuu_int);
    name = "<UUU>";
  }
  
let s0_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T0); prop = prop_of_int base });
    up = true;
    int = s0_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s0_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s0_int);
    name = "<0>";
  }
  
let s1_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T1); prop = prop_of_int base });
    up = true;
    int = s1_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s1_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s1_int);
    name = "<1>";
  }
  
let s2_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T2); prop = prop_of_int base });
    up = true;
    int = s2_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s2_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s2_int);
    name = "<2>";
  }
  
let s3_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T3); prop = prop_of_int base });
    up = true;
    int = s3_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s3_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s3_int);
    name = "<3>";
  }
  
let s4_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T4); prop = prop_of_int base });
    up = true;
    int = s4_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s4_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s4_int);
    name = "<4>";
  }
  
let s5_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T5); prop = prop_of_int base });
    up = true;
    int = s5_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s5_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s5_int);
    name = "<5>";
  }
    
let s6_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T6); prop = prop_of_int base });
    up = true;
    int = s6_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s6_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s6_int);
    name = "<6>";
  }
    
let s7_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T7); prop = prop_of_int base });
    up = true;
    int = s7_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s7_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s7_int);
    name = "<7>";
  }
    
let s8_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T8); prop = prop_of_int base });
    up = true;
    int = s8_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s8_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s8_int);
    name = "<8>";
  }
    
let s9_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_size (tdoc, T9); prop = prop_of_int base });
    up = true;
    int = s9_int;
    at_least = sizes;
    on_force_stop = 0;
    base_shared = sizes - s9_int;
    change_base = (fun base -> (base land (lnot sizes)) lor s9_int);
    name = "<9>";
  }

    
let cr_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Red); prop = prop_of_int base });
    up = true;
    int = cr_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cr_int;
    change_base = (fun base -> (base land (lnot colors)) lor cr_int);
    name = "<r>";
  }
  
let cg_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Green); prop = prop_of_int base });
    up = true;
    int = cg_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cg_int;
    change_base = (fun base -> (base land (lnot colors)) lor cg_int);
    name = "<g>";
  }
  
let cb_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Blue); prop = prop_of_int base });
    up = true;
    int = cb_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cb_int;
    change_base = (fun base -> (base land (lnot colors)) lor cb_int);
    name = "<b>";
  }
  
let cc_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Cyan); prop = prop_of_int base });
    up = true;
    int = cc_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cc_int;
    change_base = (fun base -> (base land (lnot colors)) lor cc_int);
    name = "<c>";
  }
  
let cm_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Magenta); prop = prop_of_int base });
    up = true;
    int = cm_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cm_int;
    change_base = (fun base -> (base land (lnot colors)) lor cm_int);
    name = "<m>";
  }
  
let cy_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Yellow); prop = prop_of_int base });
    up = true;
    int = cy_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cy_int;
    change_base = (fun base -> (base land (lnot colors)) lor cy_int);
    name = "<y>";
  }
    
let ck_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, Black); prop = prop_of_int base });
    up = true;
    int = ck_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - ck_int;
    change_base = (fun base -> (base land (lnot colors)) lor ck_int);
    name = "<k>";
  }
    
let cw_tag  = {
    constr = (fun base tdoc -> 
        { expr = Ttag_color (tdoc, White); prop = prop_of_int base });
    up = true;
    int = cw_int;
    at_least = colors lor spc_int;
    on_force_stop = 0;
    base_shared = colors - cw_int;
    change_base = (fun base -> (base land (lnot colors)) lor cw_int);
    name = "<w>";
  }

let tags = [b_tag; i_tag; em_tag; s_tag; tt_tag; u_tag; uu_tag; uuu_tag;
    s0_tag;    
    s1_tag;    
    s2_tag;    
    s3_tag;    
    s4_tag;    
    s5_tag;    
    s6_tag;    
    s7_tag;    
    s8_tag;    
    s9_tag;    
    cr_tag;
    cg_tag;
    cb_tag;
    cc_tag;
    cm_tag;
    cy_tag;
    ck_tag;
    cw_tag;
  ]

let rec before_pl t pos =
  if t.ints.(pos) land pl_int <> 0 then
    before_pl t (pos-1)
  else (pos+1)

let diff_up p1 p2 =
  let diff = p1 lxor p2 in
  diff land p2
    
let rec find_lowcost tag t p cost pos =
  if pos = t.len then -1 else
  let prop = t.ints.(pos) in
  if prop = p then pos else
  if prop land tag.on_force_stop <> 0 then -1 else
  if diff_up p prop - diff_up t.ints.(pos-1) prop > cost then pos else
    find_lowcost tag t p cost (pos+1)

let rec find_prop_again tag t p _ pos =
  if pos = t.len then -1 else
  let prop = t.ints.(pos) in
  if prop = p then pos else
  if prop land tag.on_force_stop <> 0 then -1 else
    find_prop_again tag t p 0 (pos+1)

let imm_abort _ _ _ _ _  = -1
    
let cost_function = ref find_lowcost
    
let rec compute_range t pos tag base = 
  if pos = t.len then pos else
  let prop = t.ints.(pos) in
  if tag.at_least land prop <> 0 &&
    tag.on_force_stop land prop = 0 then
    if tag.base_shared land prop land base = 0 then
      compute_range t (pos+1) tag base
    else (* we failed because we found a property shared by the base.
    This is clearly suboptimal since it might be faster not to close the
    current tag (simply opening the root one), than closing all nested
    tags to go back to root property.
  
    In which cases should we continue ?
    - We can find later exactly the same property as the current one,
    without breaking (on_force_stop) the current one.
    *)
    let p = t.ints.(pos-1) in
    let cost = diff_up base prop - diff_up p prop - 1 in
    let pos2 = !cost_function tag t p cost (pos+1)   in
    if debug_flat then begin
        Printf.printf "FIND PROP AGAIN %d / %d for %s" pos2 t.len tag.name; print_newline ();
      end;
    if pos2 > pos then
      compute_range t (pos2+1) tag base
    else
    if prop land pl_int <> 0 &&
      t.ints.(pos-1) land pl_int <> 0 then
      before_pl t (pos-2)
    else
      pos
  else (* aucun tag ne peut s'arreter au milieu d'un PL *)
  if prop land pl_int <> 0 &&
    t.ints.(pos-1) land pl_int <> 0 then
    before_pl t (pos-2)
  else
    pos
  
let rec remove_pl t pos =
  if pos < t.len then
    let prop = t.ints.(pos) in
    if prop land pl_int <> 0 then begin
        t.ints.(pos) <- prop lxor pl_int;
        remove_pl t (pos+1) 
      end
    else pos
  else pos
      
let rec build t pos list base = 
  if debug_flat then begin
      Printf.printf "base: %d" base; print_newline ();
    end;
  if pos = t.len then
    let _ =
      if debug_flat then begin
          Printf.printf "END BUILD at %d" pos; print_newline ();
        end;
    in
    list_tdoc (List.rev list) base else
  let _ =
    if debug_flat then begin
        Printf.printf "build at %d" pos; print_newline ();
      end;
  in
  let prop = t.ints.(pos) in
  if prop land pl_int <> 0 then
    let end_pl  = remove_pl t pos in
    if debug_flat then begin
        Printf.printf "ADDING <PL> from %d to %d" pos end_pl; print_newline ();
      end;
    let t_pl = { t with len = end_pl } in
    mark_pl t_pl pos;
    let tdoc = { 
        expr = Ttag_PL (build t_pl pos [] (base land (lnot pl_ints)));
        prop = prop_of_int base
      }
    in
    build t end_pl (tdoc :: list) base
  else
  let _ =
    if debug_flat then begin
        Printf.printf "prop[%d] %d" pos prop; print_newline ();
      end;
  in
  if prop = base ||
    (is_spc prop && space_int base = space_int prop) then
(* props are compatible *)
    build t (pos+1) ((string_tdoc t.strs.(pos) prop) :: list) base
  else
  let diff = prop lxor base in
  if debug_flat then begin
      Printf.printf "NEW POS: %d / %d" pos t.len; 
      if prop land pl_int <> 0 then
        Printf.printf " which is PL";
      print_newline ();
    end;
  let diff_up = prop land diff in
  if debug_flat then begin
      Printf.printf "DIFF UP: %x" diff_up; dump_prop (prop_of_int diff_up); print_newline ();
    end;
  let diff_down = base land diff in
  if debug_flat then begin      
      Printf.printf "DIFF DOWN: %x" diff_down; dump_prop (prop_of_int diff_down); 
      print_newline ();
    end;
  let (max, tags) = List.fold_left (fun (max, tags) tag ->
        let use_tag = if tag.up then
            tag.int land diff_up <> 0
          else 
            tag.int land diff_down <> 0
        in
        if use_tag then
          let pos2 = compute_range t (pos+1) tag base in
          let range = pos2 - pos in
          if debug_flat then begin
              Printf.printf "*** *** Range of %s: %d" tag.name range; print_newline ();
            end;
          if range > max then (range, [tag]) else
          if range = max then (range, tag :: tags) else
            (max, tags)
        else (max, tags)
    ) (1,[]) tags in
  if tags = [] then
      failwith "Not yet implemented"
  else
  let tdoc = 
    let rec iter base tags =
      match tags with
        [] ->
          if debug_flat then begin
              Printf.printf "BUILD in %d-%d" pos (pos + max); print_newline ();
            end;
          build { t with len = pos + max } pos [] base
      | tag :: tags ->
          let new_base = tag.change_base base in
          if debug_flat then begin
              Printf.printf "ADDING %s" tag.name; print_newline ();
            end;
          tag.constr base (iter new_base tags)
    in
    iter base tags 
  in
  build t (pos+max) (tdoc :: list) base
    
let flat_and_build tdoc =
  let t = flatten_tdoc tdoc in
  mark_pl t 0;
  build t 0 [] (int_of_prop prop_root)
    
let subflat_and_build tdoc prop =
  if debug_flat then begin
      Printf.printf "SUBFLATTEN START"; print_newline ();
    end;
  let orig_size = B8_valid.size tdoc in
  let t = flatten_tdoc tdoc in
  mark_pl t 0;
  let tdoc2 = build t 0 [] (int_of_prop prop) in
  let size2 = B8_valid.size tdoc2 in
  if orig_size > size2 then tdoc2 else
    
    begin 
      if debug_flat then begin
          Printf.printf "SUBFLATTEN failed: %d/%d" size2 orig_size;
          print_newline ();
          Printf.printf "ORIGINAL TREE:"; print_newline ();
          B8_dump.dump_simple tdoc;
          
          print_newline ();
          Printf.printf "GENERATED TREE:"; print_newline ();
          B8_dump.dump_simple tdoc2;
          print_newline ();
          
          Printf.printf "ORIGINAL FILE:"; print_newline ();
          B8_emit.emit_simple tdoc stdout;
          
          print_newline ();
          Printf.printf "GENERATED FILE:"; print_newline ();
          B8_emit.emit_simple tdoc2 stdout;
          print_newline ();
        end;
      
      let tdoc =
        try
          cost_function := find_prop_again;
          mark_pl t 0;
          let tdoc2 = build t 0 [] (int_of_prop prop) in
          let size2 = B8_valid.size tdoc2 in
          cost_function := find_lowcost;
          if orig_size > size2 then tdoc2 else 
          let _ = () in
          cost_function := imm_abort;
          mark_pl t 0;
          let tdoc2 = build t 0 [] (int_of_prop prop) in
          let size2 = B8_valid.size tdoc2 in
          cost_function := find_lowcost;
          if orig_size > size2 then tdoc2 else tdoc
        with _ -> 
            cost_function := find_lowcost;
            tdoc
      in
      
      tdoc
    end

(********************************************************************

DIVIDE:

- Divide the document in smaller parts (< 100 blocks)
- Call FLATTEN on each part
  
  ********************************************************************)

(********************************************************************

RANDOM:

- Randomly Divide the document in smaller parts
- Call FLATTEN on each part
  
  ********************************************************************)

(********************************************************************

SUBFLATTEN:

- Call FLATTEN on small subtrees
  
  ********************************************************************)

let rec subtree tdoc =
  match tdoc.expr with
    Ttag_B tdoc2 -> (subtree tdoc2)+1
  | Ttag_EM tdoc2 ->  (subtree tdoc2)+1
  | Ttag_I tdoc2 ->  (subtree tdoc2)+1
  | Ttag_PL tdoc2 ->  
      let n = subtree tdoc2 in
      if n < 3 then raise not_found;
      tdoc.expr <- Ttag_PL (subflat_and_build tdoc2 tdoc2.prop);
      raise not_found
  | Ttag_S tdoc2 ->  (subtree tdoc2)+1
  | Ttag_TT tdoc2 ->  (subtree tdoc2)+1
  | Ttag_U tdoc2 -> 
      (subtree tdoc2)+1
      (*
      let n = subtree tdoc2 in
      if n < 3 then raise not_found;
      tdoc.expr <- Ttag_U (subflat_and_build tdoc2 tdoc2.prop);
raise not_found
*)
  | Ttag_color (tdoc2, color)  ->  
      (subtree tdoc2)+1
      (*
      let n = subtree tdoc2 in
      if n < 3 then raise not_found;
      tdoc.expr <- Ttag_color (subflat_and_build tdoc2 tdoc2.prop, color);
raise not_found
  *)
  | Ttag_size (tdoc2, size) ->  
    (subtree tdoc2)+1
    (*
      let n = subtree tdoc2 in
      if n < 3 then raise not_found;
      tdoc.expr <- Ttag_size (subflat_and_build tdoc2 tdoc2.prop, size);
raise not_found
  *)
  | Tlist list -> 
      let n, list = List.fold_left (fun (n, list) tdoc2 ->
            if n < 0 then 
              try
                let n = subtree tdoc2 in
                if n < 3 then raise not_found;
                (-1, (subflat_and_build tdoc2 tdoc2.prop) :: list)
              with _ ->
                  (-1, tdoc2 :: list)
            else
            try
              let nn =  subtree tdoc2 in
              (n+nn, tdoc2 :: list)
            with _ ->
                if n > 2 then begin
                    let tdoc = { expr = Tlist (List.rev list); prop = tdoc.prop }
                    in
                    let tdoc = subflat_and_build tdoc tdoc.prop in
                    (-1, tdoc2 :: [tdoc])
                  end else
                  (-1, tdoc2 :: list)
        )  (0, []) list in
      tdoc.expr <- Tlist (List.rev list);
      if n < 0 then raise not_found else n
  | Tstring s -> 1

let rec copy_tree tdoc =
  match tdoc.expr with
    Ttag_B tdoc2 -> { tdoc with expr = Ttag_B (copy_tree tdoc2) }
  | Ttag_EM tdoc2 ->  { tdoc with expr = Ttag_EM (copy_tree tdoc2) }
  | Ttag_I tdoc2 ->  { tdoc with expr = Ttag_I (copy_tree tdoc2) }
  | Ttag_PL tdoc2 ->  { tdoc with expr = Ttag_PL (copy_tree tdoc2) }
  | Ttag_S tdoc2 ->  { tdoc with expr = Ttag_S (copy_tree tdoc2) }
  | Ttag_TT tdoc2 ->  { tdoc with expr = Ttag_TT (copy_tree tdoc2) }
  | Ttag_U tdoc2 ->  { tdoc with expr = Ttag_U (copy_tree tdoc2) }
  | Ttag_color (tdoc2, color)  ->  
      { tdoc with expr = Ttag_color (copy_tree tdoc2, color) }
  | Ttag_size (tdoc2, size) ->  
      { tdoc with expr = Ttag_size (copy_tree tdoc2, size) }
  | Tlist [tdoc] -> copy_tree tdoc
  | Tlist list -> 
      { tdoc with expr = Tlist (List.map copy_tree list) }
  | Tstring s -> { tdoc with expr = Tstring s }
      
  
let subflatten tdoc =
  let tdoc = copy_tree tdoc in
  let tdoc =
    try let n = subtree tdoc in
      if n > 2 then
        subflat_and_build tdoc tdoc.prop
      else tdoc
    with _ -> tdoc in
(*  let tdoc = retype_tree prop_root tdoc in *)
  tdoc

  
(********************************************************************

REDONDANT:

- Randomly Divide the document in smaller parts
- Call FLATTEN on each part
  
  ********************************************************************)

let uuu_ints = u_int + uu_int + uuu_int
  
let rec redondant tdoc =
  match tdoc.expr with
    Ttag_B tdoc2 ->
      let (prop, tdoc2) as recv = redondant tdoc2 in
      if prop land b_int <> 0 then b_int lxor 
          prop, { tdoc with expr = Ttag_B tdoc2 }
      else recv
  
  | Ttag_EM tdoc2 ->
      let (prop, tdoc2) as recv = redondant tdoc2 in
      prop lxor em_int, { tdoc with expr = Ttag_EM tdoc2 }
  
  | Ttag_I tdoc2 -> 
      let (prop, tdoc2) as recv = redondant tdoc2 in
      if prop land i_int <> 0 then i_int lxor 
          prop, { tdoc with expr = Ttag_I tdoc2 }
      else recv
        
  | Ttag_PL tdoc2 -> 
      (* PL should be removed in retype_doc ? *)
      let (prop, tdoc2) as recv = redondant tdoc2 in
      prop, { tdoc with expr = Ttag_PL tdoc2 }
  
  | Ttag_S tdoc2 ->
      let (prop, tdoc2) as recv = redondant tdoc2 in
      if prop land s_int <> 0 then 
        (s_int lxor prop), { tdoc with expr = Ttag_S tdoc2 }
      else recv
  
  | Ttag_TT tdoc2 -> 
      let (prop, tdoc2) as recv = redondant tdoc2 in
      if prop land tt_int <> 0 then
        tt_int lxor prop, { tdoc with expr = Ttag_TT tdoc2 }
      else recv
  
  | Ttag_U tdoc2 -> 
      let (prop, tdoc2) as recv = redondant tdoc2 in
      if prop land uuu_ints <> 0 then begin
        (if prop land uuu_int <> 0 then uuu_int else
          if prop land uu_int <> 0 then uu_int else u_int) lxor
            prop, { tdoc with expr = Ttag_U tdoc2 } 
        end
      else recv
  
  | Ttag_color (tdoc2, color)  ->  
      let (prop, tdoc2) as recv = redondant tdoc2 in
      let i = int_of_color color in
      if prop land i <> 0 then i lxor
        prop, { tdoc with expr = Ttag_color (tdoc2,color) } 
      else recv
  
  | Ttag_size (tdoc2, size) ->  
      let (prop, tdoc2) as recv = redondant tdoc2 in
      let i = int_of_size size in
      if prop land i <> 0 then i lxor
        prop , { tdoc with expr = Ttag_size (tdoc2, size) }
      else recv
  
  | Tlist [tdoc2] -> redondant tdoc2
  | Tlist list -> 
      let prop, list = List.fold_left (fun (prop, list) tdoc2 -> 
            let prop2, tdoc2 = redondant tdoc2 in
            prop lor prop2, tdoc2 :: list) (0,[]) list in
      prop, { tdoc with expr = Tlist (List.rev list) }
  
  | Tstring s -> 
      (if  s = "" then 0 else
        let prop = int_of_prop tdoc.prop in
        if only_spaces s then
          space_int prop
        else prop), tdoc
      
let redondant tdoc =
  let _, tdoc = redondant tdoc in
  tdoc

      
let _ = 
  register "simple" simplify 20;
  register "b8_flatten" flat_and_build 40;
  register "subflatten" subflatten 30; 
  register "merge" merge 10;
  register "redondant" redondant 5
