open B8_types
open B8_prop
open B8_dump

let counter = ref 0
let simple_strings = ref false
let escape = ref false

let escaped s =
  for i = 0 to String.length s - 1 do
    match s.[i] with
      '\r' -> s.[i] <- '\n'
    | _ -> ()
  done;
  s
  
let rec output_tdoc oc tdoc =
  match tdoc.expr with
    Ttag_B tdoc -> 
      output_string oc "<B>"; output_tdoc oc tdoc; output_string oc "</B>"
  | Ttag_EM tdoc -> 
      output_string oc "<EM>"; output_tdoc oc tdoc; output_string oc "</EM>"
  | Ttag_I tdoc -> 
      output_string oc "<I>"; output_tdoc oc tdoc; output_string oc "</I>"
  | Ttag_PL tdoc -> 
      output_string oc "<PL>"; output_tdoc oc tdoc; output_string oc "</PL>"
  | Ttag_S tdoc -> 
      output_string oc "<S>"; output_tdoc oc tdoc; output_string oc "</S>"
  | Ttag_TT tdoc -> 
      output_string oc "<TT>"; output_tdoc oc tdoc; output_string oc "</TT>"
  | Ttag_U tdoc -> 
      output_string oc "<U>"; output_tdoc oc tdoc; output_string oc "</U>"
  | Ttag_color (tdoc, color) -> 
      let color = char_of_color color in
      output_char oc '<';
      output_char oc color;
      output_char oc '>';
      output_tdoc oc tdoc; 
      output_char oc '<';
      output_char oc '/';
      output_char oc color;
      output_char oc '>';
  | Ttag_size (tdoc, size) -> 
      let size = char_of_size size in
      output_char oc '<';
      output_char oc size;
      output_char oc '>';
      output_tdoc oc tdoc; 
      output_char oc '<';
      output_char oc '/';
      output_char oc size;
      output_char oc '>';
  | Tlist list ->
      output_tdocs oc list    
  | Tstring string -> output_string oc
        (if string = "" then "" else
        if !simple_strings then
          let c1 = string.[0] in
          incr counter;
          let c2 = string.[String.length string - 1] in
          (if is_space c1 then " " else "") ^
            (string_of_int !counter) ^
          (if is_space c2 then " " else "")
        else
          (if !escape then escaped string else string))

and output_tdocs oc list =
  match list with
    [] -> ()
  | tdoc :: tail ->
      output_tdoc oc tdoc;
      output_tdocs oc tail
      
  
let emit tdoc filename  = 
  let oc = open_out filename in
  output_tdoc oc tdoc;
  close_out oc

let emit_oc tdoc oc = output_tdoc oc tdoc
  
let emit_simple tdoc oc =
  simple_strings := true;
  counter := 0;
  (try emit_oc tdoc oc with _ -> ());
  simple_strings := false