let file1 = "file_in"
let file2 = "file_out"

(*let file_root = "/home/pauillac/cristal6/guesdon/public_html/icfp"*)
let full_file1 = ref (*Filename.concat file_root*) file1
let full_file2 = ref (*Filename.concat file_root*) file2
let url_root = "http://pauillac.inria.fr/~guesdon/icfp"
let url_file1 = url_root^"/"^file1
let url_file2 = url_root^"/"^file2

let validator_url = Url.url_of_string "http://cristal.inria.fr/bin/compare"

let reply_ok = "Document are equivalent"

let max_depth = ref 10
let max_width = ref 10
(* size in bytes *)
let max_size = ref 1000 
let string_size = ref 10

(* the command for the simplifier *)
let com = ref (None : string option)

(* the file to use as input instead of random generation *)
let file = ref (None : string option)

(* the options to generate spaces instead of CR, LF and TAB *)
let spaces_only = ref false

(* command line options *)
let options = [
  "-d", Arg.Int (fun n -> max_depth := n), " <n> Max depth" ;
  "-w", Arg.Int (fun n -> max_width := n), " <n> Max width" ;
  "-s", Arg.Int (fun n -> max_size := n), " <n> Max size" ;
  "-ss" , Arg.Int (fun n -> string_size := n), " <n> Random bound for string size" ;
  "-com", Arg.String (fun s -> com := Some s), "<com>  Simplifier command" ;
  "-spaces_only", Arg.Set spaces_only, " to generate spaces instead of CR, LF and TAB in random generation" ;
  "-f", Arg.String (fun s -> file := Some s), "<file>  Use file as input instead of random generation" ;
  "-f2", Arg.String (fun s -> full_file2 := s), "<file>  Use file as output" ;
] 

let _ = Arg.parse options (fun _ -> ()) 
    ("Usage : " ^ Sys.argv.(0) ^ " [options]\nOptions are:")



let test () =
  let buf = Buffer.create 64 in
  Http.get_page 
(*    (Url.url_of_string "http://pauillac.inria.fr/~guesdon/")*)
    validator_url
    [ Http.Headers [ Http.Generic ("Content-Type", "application/x-www-form-urlencoded") ] ;
      Http.Post ;
      Http.Args 
	[
	  "fileone", url_file1 ;
	  "filetwo", url_file2
	] 
    ]
    (fun s -> Buffer.add_string buf s);
  let reply = Buffer.contents buf in
  try
    let _ = Str.search_forward (Str.regexp reply_ok) reply 0 in
    true
  with
    _ ->
      prerr_endline reply ;
      false

(* Auto-testing *)
let input = "<r>  xxx </r><b> yyy  </b>"
let output = "<r> xxx  <b>  yyy </b></r>"
let f1 = Filename.temp_file "testfile" ""
let f2 = Filename.temp_file "testfile2" ""
let chanout1 = open_out f1
let chanout2 = open_out f2
let _ = output_string chanout1 input
let _ = output_string chanout2 output
let _ = close_out chanout1
let _ = close_out chanout2
let _ = try Unix.unlink f1 with _ -> ()
let _ = try Unix.unlink f2 with _ -> ()
let _ =
  if test () then
    ()
  else
    (
     prerr_endline "Autotest failed" ;
     (*exit 1*)
    )

let _ = Random.init (int_of_float (Unix.time ()))

(* Generation of file. *)
type action = 
    String of string
  | Begin of string
  | End

(* Get a random tag. *)
let random_tag () =
  match Random.int 25 with
  | 10 -> "B" 
  | 11 -> "EM"
  | 12 -> "I" 
  | 13 -> "PL"
  | 14 -> "S" 
  | 15 -> "TT"
  | 16 -> "U" 
  | 17 -> "r" 
  | 18 -> "g" 
  | 19 -> "b" 
  | 20 -> "c" 
  | 21 -> "m" 
  | 22 -> "y" 
  | 23 -> "k" 
  | 24 -> "w" 
  | n -> string_of_int n (* should be 0 .. 9 *)

(* Get a random char, among those autorized. *)
let random_char () =
  let n = Random.int 127 in
  if n >= 33 then
    match Char.chr n with
      '>' -> '_'
    | '<' -> '&'
    | c -> c
  else
    if !spaces_only then
      '\032'
    else
      match n mod 4 with
      | 0 -> '\032' (* SPC *)
      | 1 -> '\013' (* CR *)
      | 2 -> '\010' (* LF *)
      | _ -> '\009' (* TAB *)
      
(* Get a random string. *)
let random_string () =
  let size = (Random.int !string_size) + 1 in
  let s = String.create size in
  for i = 0 to (size - 1) do
    s.[i] <- random_char ()
  done;
  s

(* Randomly get the next action. *)
let rec get_action cur_size cur_depth cur_width =
  if cur_size >= !max_size then
    End
  else
    match Random.int 3 with
      0 -> 
	let s = random_string () in
	String s
    | 1 -> 
	if cur_depth >= !max_depth then
	  get_action cur_size cur_depth cur_width
	else
	  Begin (random_tag ())
    | _ -> 
	End

let generate () =
  let chanout = open_out !full_file1 in
  let tag_stack = Stack.create () in
  let cur_size = ref 0 in
  let cur_depth = ref 0 in
  let cur_width = ref 0 in
  let _ = 
    try
      while true do
	match get_action !cur_size !cur_depth !cur_width with
	  String s ->
	    cur_size := !cur_size + (String.length s);
	    output_string chanout s
	| Begin s ->
	    Stack.push s tag_stack ;
	    incr cur_depth;
	    let s2 = "<"^s^">" in
	    cur_size := !cur_size + (String.length s2);
	    output_string chanout s2
	| End ->
	    try 
	      let s = Stack.pop tag_stack in
	      decr cur_depth;
	      let s2 = "</"^s^">" in
	      cur_size := !cur_size + (String.length s2);
	      output_string chanout s2;
	      try 
		(* don't end if we're not even at max_size / 2 *)
		if !cur_size < !max_size / 2 then
		  ()
		else
		  let _ = Stack.top tag_stack in
		  ()
	      with
		Stack.Empty ->
		  raise (Failure "")
	    with
	      Stack.Empty ->
		()
      done
    with
      Failure _ ->
	flush chanout;
	close_out chanout;
  in
  ()


let main () =
  let _ = 
    match !file with
      None -> generate () 
    | Some f -> full_file1 := f
  in
  match !com with
    None -> 
      prerr_endline "You must give a command to simplify the file.";
      exit (1)
  | Some com ->
      let command = com^" < "^ !full_file1 ^" > "^ !full_file2 in
      let n = Sys.command command in
      if n = 0 then
	if test () then
	  print_string "Test Ok :-)\n"
	else
	  print_string "Test not ok :-(\n"
      else
	(
	 prerr_endline "An error occured while executing the following command:";
	 prerr_endline command ;
	 exit 1
	)

let _ = main ()
