open B8_types


(* Principe : essayer de grouper les tags successifs et *)
(* trouver des prefixes communs dans les Tlist          *)
(* modulo une notion de                                 *)
(* commutation qu' on espere pas trop con,              *)
(* pour les factoriser.                                 *)
(* C'est pas fini : pour l'instant y a une fonction qui *)
(* regarde le plus grand prefixe commun entre deux      *)
(* listes de tags.                                      *)

type tag = B | EM | I | PL | S | TT | U | Size of size | Color of color

type modif = tag list

type mdoc = { mutable mexpr : mdoc_expr ; mutable mprop : prop }
and mdoc_expr = 
  | Mstring of string
  | Mmodif of modif * mdoc
  | Mlist of mdoc list

let rec mdoc_of_tdoc doc = 
  match doc.expr with
  | Ttag_B doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(B :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ B ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_EM doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(EM :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ EM ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_I doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(I :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ I ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_PL doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(PL :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ PL ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_S doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(S :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ S ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_TT doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(TT :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ TT ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_U doc' ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif(U :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ U ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_color (doc', c) ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif((Color c) :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ (Color c) ], ldoc) ; mprop = doc.prop } 
      end
  | Ttag_size (doc', n) ->
      begin
	let ldoc = mdoc_of_tdoc doc' in
	match ldoc.mexpr with
	| Mmodif(m, ldoc') ->
	    { mexpr = Mmodif((Size n) :: m, ldoc') ; mprop = doc.prop } 
	| ldocexpr -> 
	    { mexpr = Mmodif([ (Size n) ], ldoc) ; mprop = doc.prop } 
      end
  | Tlist(docs) -> 
      { mexpr = Mlist(List.map mdoc_of_tdoc docs) ; mprop = doc.prop } 
  | Tstring s -> { mexpr = Mstring s ; mprop = doc.prop }

let modif_prop prop tag = 
  match tag with
  | B -> { prop with b = true }
  | EM -> if prop.s then prop else { prop with em = not prop.em }
  | I -> { prop with i = true }
  | PL -> 
      if prop.u != 0 || prop.s || prop.em || prop.i || prop.b || prop.tt
      then
	{ prop_root with size = prop.size ; color = prop.color }
      else prop
  | S -> if prop.s then prop else { prop with s = true; em = false }
  | TT -> if prop.tt then prop else { prop with tt = true }
  | U -> if prop.u = 3 then prop else { prop with u = prop.u + 1 }
  | Color c -> 
      if prop.color = c then prop 
      else { prop with color = c }
  | Size n ->
      if prop.size = n then prop 
      else { prop with size = n }


(* Construit la tdoc_expr en fonction du tag et du doc a foutre devant *)
let construct_tag doc = function
  | B -> Ttag_B doc
  | EM -> Ttag_EM doc
  | I -> Ttag_I doc
  | PL -> Ttag_PL doc
  | S -> Ttag_S doc
  | TT -> Ttag_TT doc
  | U -> Ttag_U doc
  | Color c -> Ttag_color(doc, c)
  | Size n -> Ttag_size(doc, n)

(* Construit un tdoc a partir d'un mdoc *)
let rec tdoc_of_mdoc mdoc =
  match mdoc.mexpr with
  | Mmodif (m, mdoc') ->
      maketags
	(tdoc_of_mdoc mdoc')
	mdoc.mprop
	m
  | _ -> assert false

(* Construit un tdoc a partir d'un tdoc et *)
(* de la liste de tags a foutre dessus     *)
and maketags doc prop = function
  | [] -> doc
  | tag :: tags ->
      let prop' = modif_prop prop tag in
      let doc' = maketags doc prop' tags in
      { expr = construct_tag doc' tag ; prop = prop }

(* Regarde si deux tags commutent *)
let is_size = function
  | Size _ -> true
  | _ -> false

let is_color = function
  | Color _ -> true
  | _ -> false

let commute tag tag' =
  match tag, tag' with
  | t, t' when t = t' -> true
  | (B | EM | I), (B | EM | I) -> true
  | Size n, t' when not (is_size t') -> true
  | Color c, t' when not (is_color t') -> true
  | t', Size n when not (is_size t') -> true
  | t', Color c when not (is_color t') -> true
  | _, _ -> false


(* Essaie de remonter un tag en premiere position *)
let rec up tag = function
  | [] -> raise Not_found
  | tag' :: tags ->
      if tag' = tag then tags
      else
        if commute tag tag' then 
	  tag' :: (up tag tags)
	else raise Not_found

(* Calcule les permutations d'une liste *)

let rec insert x n = function
  | [] -> assert (n = 0); [ x ]
  | (x1 :: xs) as l ->
      if n = 0 then
	x :: l
      else
	x1 :: (insert x (n - 1) l)

let rec insert_everywhere x = function
  | [] -> [ [ x ] ]
  | (y :: ys) as l ->
      let l' = List.map (fun l' -> y :: l') (insert_everywhere x ys) in
      (x :: l) :: l'

let rec permutations n l = 
  if n = 0 then [ [] ]
  else if n = 1 then [ l ]
  else match l with
  | x :: xs -> 
      let ls = permutations (n - 1) xs in
      List.flatten (List.map (insert_everywhere x) ls)
  | _ -> assert false

let permutations l = permutations (List.length l) l

(* Regarde le plus long prefixe commun a deux listes de tags *)

let commons tags1 tags2 =
  List.find_all (fun tag -> List.mem tag tags2) tags1

let try_common tag tags1 tags2 =
    let tltags1 = up tag tags1 in
    let tltags2 = up tag tags2 in
    (tltags1, tltags2)

let rec build_prefixes yet commons tags1 tags2 =
  match commons with
  | [] -> List.length yet, yet, tags1, tags2
  | tag :: commons ->
      try
	let tltags1, tltags2 = try_common tag tags1 tags2 in
	build_prefixes (tag :: yet) commons tltags1 tltags2
      with
	_ -> build_prefixes yet commons tags1 tags2

let proj_1_4 (x, _, _, _) = x

let rec best = function
  | [] -> raise Not_found
  | [ r ] -> r
  | r :: res -> 
      let r' = best res in
      if (proj_1_4 r') > (proj_1_4 r) 
      then r' 
      else r

let longest_prefix tags1 tags2 =
  let lists_of_commons = permutations (commons tags1 tags2) in
  let results = 
    List.map
      (fun commons -> build_prefixes [] commons tags1 tags2)
      lists_of_commons
  in
  best results

(* Prend une liste de listes de tags et cherche le plus grand prefixe commun *)
(* entre deux consecutives *)
(* Prochaine fois *)
