open B8_types

(* supprime les espaces clairement inutiles *)
let delete_white s =
  let len = (String.length s) in
  let res = Buffer.create len in
  let rec aux i spc =
    if i < len
    then
      match s.[i] with
      | ' ' | '\r' | '\n' | '\t' ->
          if spc
          then (* forget multiple spaces *)
            aux (i+1) true
          else
            (Buffer.add_char res ' ';
             aux (i+1) true)
      | x ->
          Buffer.add_char res x; aux (i+1) false in
  aux 0 false;
  Buffer.contents res

let convert_spaces s =
  for i = 0 to String.length s - 1 do
    match s.[i] with
      '\r' -> s.[i] <- ' '
    |  _ -> ()
  done
  
(* rajoute des etiquettes a chaque noeud d'un arbre *)
let rec transf prop = function
  | Dtag_B d ->
      if prop.b
      then transf prop d
      else
        (let np = { prop with b = true } in
        { expr = Ttag_B (transf np d); prop = prop })
  | Dtag_EM d ->
      if prop.s
      then transf prop d
      else
        (let np = { prop with em = not prop.em } in
        { expr = Ttag_EM (transf np d); prop = prop })
  | Dtag_I d ->
      if prop.i
      then transf prop d
      else
        (let np = { prop with i = true } in
        { expr = Ttag_I (transf np d); prop = prop })
  | Dtag_PL d ->
      if prop.u != 0 || prop.s || prop.em || prop.i || prop.b || prop.tt
      then
        (let np = { prop_root with size = prop.size ; color = prop.color } in
        { expr = Ttag_PL (transf np d); prop = prop })
      else transf prop d
  | Dtag_S d ->
      if prop.s
      then transf prop d
      else
        (let np = { prop with s = true; em = false } in
        { expr = Ttag_S (transf np d); prop = prop })
  | Dtag_TT d ->
      if prop.tt
      then transf prop d
      else
        (let np = { prop with tt = true } in
        { expr = Ttag_TT (transf np d); prop = prop })
  | Dtag_U d ->
      if prop.u == 3
      then transf prop d
      else
        (let np = { prop with u = prop.u+1 } in
        { expr = Ttag_U (transf np d); prop = prop })
  | Dtag_color (d, c) ->
      if prop.color == c
      then transf prop d
      else
        (let np = { prop with color = c } in
        { expr = Ttag_color (transf np d, c); prop = prop })
  | Dtag_size (d, t) ->
      if prop.size == t
      then transf prop d
      else
        (let np = { prop with size = t } in
        { expr = Ttag_size (transf np d, t); prop = prop })
  | Dlist l -> { expr = Tlist (List.map (transf prop) l); prop = prop }
  | Dstring s ->
      if prop.tt
      then { expr = Tstring s; prop = prop }
      else { expr = Tstring (delete_white s); prop = prop }

let typedoc d = transf prop_root d

let rec retype_tdoc prop tdoc =
  match tdoc.expr with
    Ttag_B tdoc2 -> 
      { expr = Ttag_B (retype_tdoc { prop with b = true } tdoc2); 
        prop = prop }
  | Ttag_EM tdoc2 ->  
      if prop.s then
        retype_tdoc prop tdoc2
      else
        { 
          expr = Ttag_EM (retype_tdoc { prop with em = not prop.em } tdoc2); 
          prop = prop }
  | Ttag_I tdoc2 ->  { 
        expr = Ttag_I (retype_tdoc { prop with i = true } tdoc2); 
        prop = prop }
  | Ttag_PL tdoc2 ->       
      if prop.u != 0 || prop.s || prop.em || prop.i || prop.b || prop.tt
      then
        { 
          expr = Ttag_PL (retype_tdoc { prop with
              u = 0; s= false; em = false; i = false; b = false;  tt = false
            } tdoc2);
          prop = prop }
      else
        retype_tdoc prop tdoc2
  | Ttag_S tdoc2 ->  { 
        expr = Ttag_S (retype_tdoc { prop with em =false; s = true } tdoc2); 
        prop = prop }
  | Ttag_TT tdoc2 ->  { 
        expr = Ttag_TT (retype_tdoc { prop with tt = true } tdoc2); 
        prop = prop }
  | Ttag_U tdoc2 ->  { 
        expr = Ttag_U (retype_tdoc { prop with u = min (prop.u + 1) 3 } tdoc2); 
        prop = prop }
  | Ttag_color (tdoc2, color)  ->  
      { expr = Ttag_color (retype_tdoc { prop with color = color } tdoc2, color); 
        prop = prop }
  | Ttag_size (tdoc2, size) ->  
      { expr = Ttag_size (retype_tdoc  { prop with size = size } tdoc2, size); 
        prop = prop }
  | Tlist [tdoc] -> retype_tdoc prop tdoc
  | Tlist list -> 
      { expr = Tlist (List.map (retype_tdoc prop) list); 
        prop = prop }
  | Tstring s -> { expr = Tstring s; prop = prop }

let retype_tdoc tdoc = retype_tdoc prop_root tdoc